# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

import random
import subprocess

from dns.rdtypes.IN.AAAA import AAAA
import pytest

from aiodnsprox.config import Config

TEST_HOSTNAME = "example.org"
TEST_ADDRESS = "2001:db8::1"


@pytest.fixture
def config():
    conf = Config()
    conf._sections.clear()  # pylint: disable=protected-access
    yield conf
    conf._sections.clear()  # pylint: disable=protected-access


@pytest.fixture
def dns_server(tmp_path):
    config = tmp_path / "dnsmasq.conf"
    config.write_text("")
    port = random.randint(1 << 2, 0xFF) << 8 | 53
    proc = subprocess.Popen(
        [
            "dnsmasq",
            "-C",
            str(config),
            "-k",
            "-p",
            str(port),
            f"--host-record={TEST_HOSTNAME},{TEST_ADDRESS}",
        ]
    )
    while True:
        try:
            subprocess.check_call(f"ss -ulpn | grep -q {port}", shell=True)
        except subprocess.CalledProcessError:
            continue
        else:
            break
    yield {
        "host": "::1",
        "port": port,
        "req_hostname": TEST_HOSTNAME,
        "resp_rtype": AAAA,
        "resp_address": TEST_ADDRESS,
    }
    proc.kill()
    proc.wait()



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

"""Implementation of the proxying side of the DNS proxy."""

import asyncio
import abc
import enum
import logging
import socket
import typing
import time

import dns.asyncquery
import dns.entropy
import dns.exception
import dns.message


class DNSTransport(enum.Enum):
    """Type to identify the server proxied via :py:class:`DNSUpstream`.

    - :py:attr:`TCP` for DNS over TCP
    - :py:attr:`UDP` for DNS over UDP
    - :py:attr:`UDP_TCP_FALLBACK` for DNS over UDP with
      a fallback to DNS over TCP in case the DNS over UDP response is truncated
    """

    UDP = 0
    UDP_TCP_FALLBACK = 1
    TCP = 2


class DNSUpstream:
    """Implementation of the DNS client towards the proxied DNS server

    :param host: Host of the proxied DNS server
    :type host: str
    :param port: (Optional) port of the proxied DNS server. If no port is
                 provided, the default of the selected ``transport``
                 will be used (e.g. 53 for :py:attr:`DNSTransport.TCP` or
                 :py:attr:`DNSTransport.UDP`).
    :type port: int
    :param transport: (Optional) transport used to communicate with the proxied
                      DNS server. If no transport is provided,
                      :py:attr:`DNSTransport.UDP` will be used.
    :type transport: DNSTransport
    """

    _QUERY_FUNC: dict[DNSTransport, typing.Callable] = {
        DNSTransport.UDP: dns.asyncquery.udp,
        DNSTransport.UDP_TCP_FALLBACK: dns.asyncquery.udp_with_fallback,
        DNSTransport.TCP: dns.asyncquery.tcp,
    }
    DEFAULT_LIFETIME = 5.0
    DEFAULT_TIMEOUT = 2.0

    def __init__(
        self,
        host: str,
        port: typing.Optional[int] = None,
        transport: DNSTransport = DNSTransport.UDP,
    ):
        self._host = host
        if port is None:
            if transport in [
                DNSTransport.UDP,
                DNSTransport.UDP_TCP_FALLBACK,
                DNSTransport.TCP,
            ]:
                self._port = 53
            else:
                raise ValueError(f"Unsupported transport {transport}")
        else:
            self._port = port
        self._transport = transport
        self._query_func = self._QUERY_FUNC[transport]

    @property
    def port(self):
        """Port of the proxied DNS server

        :type: int
        """
        return self._port

    def _compute_timeout(self, start, lifetime=None):
        # https://dnspython.readthedocs.io/en/stable/_modules/dns/resolver.html
        # Resolver._compute_timeout()
        if lifetime is None:
            lifetime = self.DEFAULT_LIFETIME
        now = time.time()
        duration = now - start
        if duration < 0:  # pragma: no cover
            if duration < -1:
                # Time going backwards is bad. Just give up.
                raise dns.exception.Timeout(timeout=duration)
            # Time went backwards, but only a little.  This can
            # happen, e.g. under vmware with older linux kernels.
            # Pretend it didn't happen.
            now = start
        if duration >= lifetime:
            raise dns.exception.Timeout(timeout=duration)  # pragma: no cover
        return min(lifetime - duration, self.DEFAULT_TIMEOUT)

    @staticmethod
    def _resp_servfail(query):
        resp = dns.message.make_response(query, recursion_available=True)
        resp.set_rcode(dns.rcode.SERVFAIL)
        return resp

    async def query(
        self, query: bytes, timeout: typing.Optional[float] = None
    ) -> bytes:
        """Query proxied DNS server.

        :param query: DNS query in the on-the-wire format
        :type query: bytes
        :param timeout: (Optional) timeout for the DNS query operation. If not
                        provided and the transport to the server is
                        :py:attr:`DNSTransport.UDP`,
                        :py:attr:`DNSUpstream.DEFAULT_LIFETIME` will be used.
        :type timeout: float
        """
        qry = dns.message.from_wire(query)
        start = time.time()
        id_ = qry.id
        tuple_resp = self._transport in [DNSTransport.UDP_TCP_FALLBACK]
        if qry.id == 0:
            id_ = dns.entropy.random_16()
            qry.id = id_
        if self._transport in [DNSTransport.UDP]:
            resp = None
            lifetime = timeout
            while resp is None:
                timeout = self._compute_timeout(start, lifetime)
                try:
                    resp = await self._query_func(
                        qry, where=self._host, port=self._port, timeout=timeout
                    )
                except dns.exception.DNSException:
                    resp = self._resp_servfail(qry)
                    break
        else:
            try:
                resp = await self._query_func(
                    qry, where=self._host, port=self._port, timeout=timeout
                )
            except (dns.exception.DNSException, ConnectionRefusedError):
                resp = self._resp_servfail(qry)
                tuple_resp = False
        if tuple_resp:
            resp = resp[0]
        resp.id = id_
        return resp.to_wire()


class MockDNSUpstream(DNSUpstream):
    """Mocks an upstream by statically responding with a preconfigured set of
    records.

    :param IN: Records for the RDATA class IN. A mapping that maps the name of
               the record to its data. Currently supported are ``A`` and
               ``AAAA`` records.
    :type IN: dict
    """

    _AF = {
        "A": socket.AF_INET,
        "AAAA": socket.AF_INET6,
    }

    def __init__(self, *args, IN=None, **kwargs):
        # pylint: disable=invalid-name,unused-argument,super-init-not-called
        self._IN = {}
        if isinstance(IN, dict):
            for key in IN:
                if isinstance(IN[key], str):
                    try:
                        self._IN[key] = socket.inet_pton(self._AF[key], IN[key])
                    except OSError as exc:
                        raise ValueError(
                            f"{IN[key]} not a valid {key} record."
                        ) from exc
                elif isinstance(IN[key], bytes):
                    try:
                        socket.inet_ntop(self._AF[key], IN[key])
                    except ValueError as exc:
                        raise ValueError(
                            f"{IN[key]} not a valid {key} record."
                        ) from exc
                    self._IN[key] = IN[key]
                else:
                    raise TypeError(f"IN[{key}] of invalid type " f"{type(IN[key])}")

    async def query(
        self, query: bytes, timeout: typing.Optional[float] = None
    ) -> bytes:
        qry = dns.message.from_wire(query)
        resp = dns.message.make_response(qry, recursion_available=True)
        questions = resp.sections[dns.message.QUESTION]
        answers = resp.sections[dns.message.ANSWER]
        for question in questions:
            data = None
            if question.rdclass == dns.rdataclass.IN:
                if "A" in self._IN and question.rdtype == dns.rdatatype.A:
                    data = self._IN["A"]
                elif "AAAA" in self._IN and question.rdtype == dns.rdatatype.AAAA:
                    data = self._IN["AAAA"]
            if data is not None:
                answers.append(
                    dns.rrset.from_rdata_list(
                        question.name,
                        300,
                        [
                            dns.rdata.GenericRdata(
                                question.rdclass, question.rdtype, data
                            )
                        ],
                    )
                )
        return resp.to_wire()


class DNSUpstreamServerMixin(abc.ABC):
    """Mixin for the serving side of the proxy for easy access towards the
    proxied side.

    :param dns_upstream: The proxied DNS server.
    :type dns_upstream: :py:class:`DNSUpstream`
    :param timeout: (Optional) timeout for queries towards ``dns_upstream``.
    :type timeout: float
    """

    logger = logging.getLogger(".".join([__module__, __name__]))

    # pylint: disable=too-few-public-methods
    def __init__(
        self, dns_upstream: DNSUpstream, timeout: typing.Optional[float] = None
    ):
        self._dns_upstream = dns_upstream
        self._timeout = timeout

    async def _get_query_response(self, query, requester):
        self.logger.debug(
            "Received query from %s:\n  %s",
            requester,
            str(dns.message.from_wire(query)).replace("\n", "\n  "),
        )
        resp = await self._dns_upstream.query(query, timeout=self._timeout)
        self.logger.debug(
            "Sending response to %s:\n  %s",
            requester,
            str(dns.message.from_wire(resp)).replace("\n", "\n  "),
        )
        self.send_response_to_requester(resp, requester)

    def dns_query_received(self, query: bytes, requester):
        """The serving end of the proxy notifies that it received a DNS query
        and sends it to the proxied DNS server. When a response is received
        asynchronously, :py:meth:`send_response_to_requester` is called to
        notify the serving end about the received response.

        :param query: The DNS query in on-the-wire format to send to the
                      proxied DNS server.
        :type query: bytes
        :param requester: Identifier for the endpoint that originally requested
                          the query.
        """
        loop = asyncio.get_event_loop()
        coroutine = self._get_query_response(query, requester)
        loop.create_task(coroutine)

    @abc.abstractmethod
    def send_response_to_requester(self, response: bytes, requester):
        """Called when proxied DNS server responded to a DNS query send by
        :py:meth:`dns_query_received`.

        :param response: The DNS response in on-the-wire format received from
                         the proxied DNS server.
        :param requester: Identifier for the endpoint that originally requested
                          the query. This will have the same value as the
                          ``requester`` parameter of
                          :py:meth:`dns_query_received` for the ``query`` that
                          ``response`` is the response to.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

"""DNS over UDP serving side of the proxy."""

from .dns_server import BaseServerFactory, BaseDNSServer
from .dns_upstream import DNSUpstreamServerMixin


class DNSOverUDPServerFactory(BaseServerFactory):
    """Factory to create DNS over UDP servers"""

    # pylint: disable=too-few-public-methods
    DNS_PORT = 53

    class DNSOverUDPServer(BaseDNSServer, DNSUpstreamServerMixin):
        """DNS over UDP server implementation

        :param factory: The factory that created the DNS over DTLS server.
        :type factory: :py:class:`DNSOverUDPServerFactory`
        """

        def __init__(self, factory):
            super().__init__(dns_upstream=factory.dns_upstream)
            self.transport = None

        def send_response_to_requester(self, response, requester):
            self.transport.sendto(response, requester)

        def connection_made(self, transport):
            # pylint: disable=line-too-long
            """See `connection_made()`_

            .. _`connection_made()`: https://docs.python.org/3/library/asyncio-protocol.html#asyncio.BaseProtocol.connection_made
            """  # noqa: E501
            self.transport = transport

        def datagram_received(self, data, addr):
            # pylint: disable=line-too-long
            """See `datagram_received()`_

            .. _`datagram_received()`: https://docs.python.org/3/library/asyncio-protocol.html#asyncio.DatagramProtocol.datagram_received
            """  # noqa: E501
            self.dns_query_received(data, addr)

        def error_received(self, exc):
            # pylint: disable=line-too-long
            """See `error_received()`_

            .. _`error_received()`: https://docs.python.org/3/library/asyncio-protocol.html#asyncio.DatagramProtocol.error_received
            """  # noqa: E501
            raise exc  # pragma: no cover

        def connection_lost(self, exc):
            # pylint: disable=line-too-long,unnecessary-pass
            """See `connection_lost()`_

            .. _`connection_lost()`: https://docs.python.org/3/library/asyncio-protocol.html#asyncio.BaseProtocol.connection_lost
            """  # noqa: E501
            pass  # pragma: no cover

        async def close(self):
            if self.transport is not None:
                self.transport.close()
                self.transport = None

    def _create_server_protocol(self, *args, **kwargs):
        return self.DNSOverUDPServer(self, *args, **kwargs)

    async def create_server(self, loop, *args, local_addr=None, **kwargs):
        """Creates an :py:class:`DNSOverUDPServer` object.

        :param loop: the asyncio event loop the server should run in
        :type loop: :py:class:`asyncio.AbstractEventLoop`
        :param local_addr: A tuple for the created server to bind to. The first
                           element is the host part, the second element the
                           port.
        :type local_addr: :py:class:`typing.Tuple[str, int]`

        :returns: An :py:class:`DNSOverUDPServer` object representing an
                  DNS over DTLS server.
        :rtype: :py:class:`DNSOverUDPServer`
        """
        if local_addr is None:
            local_addr = ("localhost", self.DNS_PORT)
        if local_addr[1] is None:
            local_addr = (local_addr[0], self.DNS_PORT)

        _, protocol = await loop.create_datagram_endpoint(
            self._create_server_protocol,
            *args,
            local_addr=local_addr,
            **kwargs,
        )
        return protocol



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

import logging
import sys

import pytest
import pytest_asyncio

from aiodnsprox.cli import proxy

from .fixtures import config


@pytest_asyncio.fixture
async def servers():
    yield proxy.servers
    await proxy.close_servers()


def test_loglevel():
    assert proxy.loglevel(str(logging.CRITICAL)) == logging.CRITICAL
    assert proxy.loglevel(str(logging.ERROR)) == logging.ERROR
    assert proxy.loglevel(str(logging.WARNING)) == logging.WARNING
    assert proxy.loglevel(str(logging.INFO)) == logging.INFO
    assert proxy.loglevel(str(logging.DEBUG)) == logging.DEBUG
    assert proxy.loglevel(str(logging.NOTSET)) == logging.NOTSET
    assert proxy.loglevel("CRITICAL") == logging.CRITICAL
    assert proxy.loglevel("critical") == logging.CRITICAL
    assert proxy.loglevel("ERROR") == logging.ERROR
    assert proxy.loglevel("WARNING") == logging.WARNING
    assert proxy.loglevel("INFO") == logging.INFO
    assert proxy.loglevel("DEBUG") == logging.DEBUG
    assert proxy.loglevel("NOTSET") == logging.NOTSET
    with pytest.raises(ValueError):
        proxy.loglevel("foobar")
    with pytest.raises(ValueError):
        proxy.loglevel("1236312576465")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "argv", [[sys.argv[0], "-u"], [sys.argv[0], "-U", "localhost"]]
)
async def test_sync_main__runtime_error(monkeypatch, servers, config, argv):
    monkeypatch.setattr(sys, "argv", argv)
    with pytest.raises(RuntimeError):
        await proxy.main()
    assert not servers


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "argv",
    [
        [sys.argv[0], "-U", "9.9.9.9", "-u", "localhost", "53", "more"],
        [sys.argv[0], "-U", "udp", "9.9.9.9", "53", "more", "-u"],
    ],
)
async def test_sync_main__value_error(monkeypatch, servers, config, argv):
    monkeypatch.setattr(sys, "argv", argv)
    with pytest.raises(ValueError):
        await proxy.main()
    assert not config
    assert not servers


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "argv, mock_upstream, exp_transports, exp_credentials",
    [
        ([sys.argv[0], "-U", "9.9.9.9", "-u"], False, 1, None),
        ([sys.argv[0], "-U", "9.9.9.9", "-u", "localhost"], False, 1, None),
        ([sys.argv[0], "-U", "9.9.9.9", "-u", "localhost", "5300"], False, 1, None),
        ([sys.argv[0], "-U", "9.9.9.9", "53", "-u"], False, 1, None),
        ([sys.argv[0], "-U", "tcp", "9.9.9.9", "53", "-u"], False, 1, None),
        ([sys.argv[0], "-U", "9.9.9.9", "-u", "-d"], False, 0, None),
        ([sys.argv[0], "-U", "9.9.9.9", "-u", "-c"], False, 0, None),
        (
            [
                sys.argv[0],
                "-U",
                "9.9.9.9",
                "-u",
                "-d",
                "--dtls-credentials",
                "client_identifier",
                "secretPSK",
            ],
            False,
            2,
            {
                "client_identity": "client_identifier",
                "psk": "secretPSK",
            },
        ),
        (
            [
                sys.argv[0],
                "-U",
                "9.9.9.9",
                "-u",
                "-d",
                "-c",
                "--dtls-credentials",
                "client_identifier",
                "secretPSK",
            ],
            False,
            3,
            {
                "client_identity": "client_identifier",
                "psk": "secretPSK",
            },
        ),
        ([sys.argv[0], "-U", "9.9.9.9", "-u", "-C", "test.yaml"], False, 1, None),
        (
            [
                sys.argv[0],
                "-U",
                "9.9.9.9",
                "-u",
                "-d",
                "--dtls-credentials",
                "client_identifier",
                "secretPSK",
            ],
            False,
            2,
            {
                "client_identity": "client_identifier",
                "psk": "secretPSK",
            },
        ),
        (
            [
                sys.argv[0],
                "-U",
                "9.9.9.9",
                "-u",
                "-c",
                "--dtls-credentials",
                "client_identifier",
                "secretPSK",
            ],
            False,
            2,
            {
                "client_identity": "client_identifier",
                "psk": "secretPSK",
            },
        ),
        ([sys.argv[0], "-U", "9.9.9.9", "-u", "-C", "test.yaml"], True, 1, None),
        ([sys.argv[0], "-U", "9.9.9.9", "-u", "-C", "test.yaml"], True, 1, None),
    ],
)
async def test_sync_main__success(
    monkeypatch,
    mocker,
    servers,
    config,
    argv,
    mock_upstream,
    exp_transports,
    exp_credentials,
):
    monkeypatch.setattr(sys, "argv", argv)
    assert not mock_upstream or "-C" in argv
    if "-C" in argv:
        if mock_upstream:
            mocker.patch(
                "argparse.open",
                mocker.mock_open(
                    read_data="""
test: foobar
mock_dns_upstream: {}"""
                ),
            )
        else:
            mocker.patch("argparse.open", mocker.mock_open(read_data="test: foobar"))
    # override default ports so we can run tests as non-root
    monkeypatch.setattr(
        proxy.HostPortAction,
        "DEFAULT_PORTS",
        {
            "dtls": 5853,
            "udp": 5300,
            "coap": None,
        },
    )
    if ("-c" in argv or "-d" in argv) and "--dtls-credentials" not in argv:
        with pytest.raises(RuntimeError):
            await proxy.main()
    else:
        await proxy.main()
        assert len(config["transports"]) == exp_transports
        assert config.get("dtls_credentials") == exp_credentials
        assert len(servers) == exp_transports
        if "-C" in argv:
            assert config["test"] == "foobar"


@pytest.mark.asyncio
async def test_sync_main__success_pre_config(monkeypatch, config):
    monkeypatch.setattr(sys, "argv", [sys.argv[0], "-U", "9.9.9.9", "-u"])
    monkeypatch.setattr(
        proxy.HostPortAction,
        "DEFAULT_PORTS",
        {
            "dtls": 5853,
            "udp": 5300,
            "coap": None,
        },
    )
    await proxy.main(config={"test": "foobar"})
    assert config["test"] == "foobar"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

"""A datagram-based DNS-over-X proxy."""

import argparse
import asyncio
import logging
import typing

from aiodnsprox.config import Config
from aiodnsprox import coap
from aiodnsprox import dns_upstream
from aiodnsprox import dtls
from aiodnsprox import udp


FACTORIES = {
    "coap": coap.DNSOverCoAPServerFactory,
    "dtls": dtls.DNSOverDTLSServerFactory,
    "udp": udp.DNSOverUDPServerFactory,
}
servers: list[typing.Tuple[asyncio.Transport, asyncio.Protocol]] = []


def loglevel(value):
    """Type function to set the log level via command-line.

    :param value: either a name for a log level or their numeric representation
                  (as string).
    :type: str
    :returns: The (integer) log level.
    """
    if value.isdigit():
        level = int(value)
        if logging.getLevelName(level).startswith("Level "):
            raise ValueError(f'Invalid log level "{value}"')
    else:
        level = logging.getLevelName(value.upper())
        if isinstance(level, str):
            raise ValueError(f'Invalid log level "{value}"')
    logging.basicConfig(level=level)
    return level


class HostPortAction(argparse.Action):  # pylint: disable=R0903
    """Argument parser Action class to parse a (host, port) pair into a
    dictionary."""

    DEFAULT_HOST = "localhost"
    DEFAULT_PORTS = {
        "coap": None,
        "dtls": dtls.DNSOverDTLSServerFactory.DODTLS_PORT,
        "udp": udp.DNSOverUDPServerFactory.DNS_PORT,
    }

    def __call__(self, parser, namespace, values, option_string, *args, **kwargs):
        if len(values) > 2:
            raise ValueError(f"{option_string} requires 2 or less arguments")
        if len(values) == 0:
            host = self.DEFAULT_HOST
            port = self._default_port()
        else:
            host = values[0]
            if len(values) == 1:
                port = self._default_port()
            else:
                port = int(values[1])
        if not hasattr(namespace, "transports"):
            setattr(namespace, "transports", argparse.Namespace())
        setattr(namespace.transports, self.dest, {"host": host, "port": port})

    def _default_port(self):
        return self.DEFAULT_PORTS[self.dest]


class DTLSCredentialsAction(argparse.Action):  # pylint: disable=R0903
    """Argument parser Action class to parse PSK DTLS credentials into a
    dictionary."""

    def __call__(self, parser, namespace, values, option_string, *args, **kwargs):
        setattr(
            namespace,
            self.dest,
            {
                "client_identity": values[0],
                "psk": values[1],
            },
        )


class UpstreamAction(HostPortAction):  # pylint: disable=R0903
    """Argument parser Action class to parse upstream parameters into a
    dictionary."""

    TRANSPORTS = {
        "tcp": dns_upstream.DNSTransport.TCP,
        "udp": dns_upstream.DNSTransport.UDP,
        "udp+tcp": dns_upstream.DNSTransport.UDP_TCP_FALLBACK,
    }

    def __call__(self, parser, namespace, values, option_string, *args, **kwargs):
        if len(values) > 3:
            raise ValueError(f"{option_string} requires 3 or less arguments")
        if len(values) == 3:
            transport = self._get_transport(values[0])
            host = values[1]
            port = int(values[2])
        else:
            transport = dns_upstream.DNSTransport.UDP
            host = values[0]
            if len(values) == 2:
                port = int(values[1])
            else:
                port = None
        setattr(
            namespace, self.dest, {"host": host, "port": port, "transport": transport}
        )

    def _get_transport(self, val):
        return self.TRANSPORTS[val]


def build_argparser():
    """Build argument parser for the proxy CLI command."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-v", "--verbosity", type=loglevel, default="WARNING", help="Sets the log level"
    )
    parser.add_argument(
        "-C",
        "--config-file",
        type=argparse.FileType("r", encoding="utf-8"),
        help="Config YAML file",
    )
    group = parser.add_argument_group("Transports")
    group.add_argument(
        "-u",
        "--udp",
        nargs="*",
        action=HostPortAction,
        default=None,
        metavar="host [port]",
        help="Start DNS-over-UDP proxy",
    )
    group.add_argument(
        "-d",
        "--dtls",
        nargs="*",
        action=HostPortAction,
        default=None,
        metavar="host [port]",
        help="Start DNS-over-DTLS proxy",
    )
    group.add_argument(
        "-c",
        "--coap",
        nargs="*",
        action=HostPortAction,
        default=None,
        metavar="host [port]",
        help="Start DNS-over-CoAP proxy",
    )
    parser.add_argument(
        "--dtls-credentials",
        nargs=2,
        action=DTLSCredentialsAction,
        default=None,
        metavar="client_id psk",
        help="DTLS credentials",
    )
    parser.add_argument(
        "-U",
        "--upstream-dns",
        action=UpstreamAction,
        default=None,
        nargs="+",
        metavar="[{udp,tcp,udp+tcp}] host [port]",
        help="Upstream server configuration. Required either "
        "as CLI argument or via config file. udp+tcp "
        "tries to use UDP first, then falls back to TCP.",
    )
    return parser


def get_config(args, pre_config: typing.Optional[typing.Mapping] = None):
    """Get :py:class:`aiodnsprox.config.Config` object from CLI arguments and
    configuration file as provided in the CLI arguments (CLI arguments take
    precedence over file configuration).

    :param args: Parsed CLI arguments.
    :type args: :py:class:`argparse.Namespace`
    """
    config = Config()
    if pre_config is not None:
        config.add_config(pre_config)
    if args.config_file:
        config.add_yaml_config(args.config_file)
        args.config_file.close()
    del args.config_file
    config.add_args_config(args)
    if "upstream_dns" not in config and "mock_dns_upstream" not in config:
        raise RuntimeError("No upstream DNS server provided")
    if "transports" not in config:
        raise RuntimeError("No proxy config provided")
    return config


def get_factory(upstream, transport):
    """Get server factory for transport.

    :param upstream: The proxied DNS server.
    :type upstream: :py:class:`aiodnsprox.dns_upstream.DNSUpstream`
    :param transport: One of the following strings:

        - ``"udp"``: Gets a DNS over UDP server factory.
        - ``"dtls"``: Gets a DNS over DTLS server factory.
        - ``"coap"``: Gets a DNS over CoAP server factory.
    :type transport: str
    """
    return FACTORIES[transport](upstream)


async def close_servers():
    """Close all servers that got started by :py:func:`main`."""
    for server in list(servers):
        await server.close()
        servers.remove(server)


async def main(config=None):
    """Asynchronous entry point.

    Parses CLI parameters and creates a config from it. Starts all serving DNS
    servers and sets up the upstream cliend towards the proxied DNS server."""
    parser = build_argparser()
    args = parser.parse_args()
    config = get_config(args, pre_config=config)

    loop = asyncio.get_event_loop()
    if "mock_dns_upstream" in config:
        upstream = dns_upstream.MockDNSUpstream(**config["mock_dns_upstream"])
    else:
        upstream = dns_upstream.DNSUpstream(**config["upstream_dns"])
    for transport, args in config["transports"].items():
        factory = get_factory(upstream, transport)
        servers.append(
            await factory.create_server(
                loop=loop, local_addr=(args["host"], args["port"])
            )
        )


def sync_main():  # pragma: no cover
    """Synchronous entry point."""
    asyncio.Task(main())
    loop = asyncio.get_event_loop()
    try:
        loop.run_forever()
    finally:
        asyncio.ensure_future(close_servers(), loop=loop)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

import argparse
import io
import pprint

import pytest

from aiodnsprox.config import Config

from .fixtures import config


def test_config__empty(config):
    assert not config
    assert "test" not in config
    with pytest.raises(KeyError):
        assert config["test"]
    assert config.get("test") is None
    assert config.get("test", "foobar") == "foobar"
    assert str(config) == "{}"
    assert repr(config) == "<Config: {}>"


def test_config__add_config(config):
    conf2 = Config()  # also test singleton/borg behavior here
    assert not config
    assert not conf2
    assert "test" not in config
    assert "test" not in conf2
    config.add_config({"test": "foobar"})
    assert "test" in config
    assert "test" in conf2
    assert config["test"] == "foobar"
    assert conf2["test"] == "foobar"
    assert config.get("test") == "foobar"
    assert conf2.get("test") == "foobar"
    assert str(config) == pprint.pformat({"test": "foobar"})
    assert str(conf2) == pprint.pformat({"test": "foobar"})


def test_config__add_yaml_config(config):
    yaml_file = io.StringIO("test: foobar\n")
    config.add_yaml_config(yaml_file)
    assert config["test"] == "foobar"


def test_config__add_args_config(config):
    args = argparse.Namespace(
        test="foobar", a_namespace=argparse.Namespace(snafu=42), nothing=None
    )
    config.add_args_config(args)
    assert config["test"] == "foobar"
    assert config["a_namespace"] == {"snafu": 42}
    assert "nothing" not in config



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

"""DNS over CoAP serving side of the proxy."""

import os
import asyncio
import base64

import aiocoap                              # type: ignore
import aiocoap.resource                     # type: ignore
import aiocoap.transports.tinydtls_server   # type: ignore
from aiocoap.numbers import NOT_ACCEPTABLE  # type: ignore

from .config import Config
from .dns_server import BaseServerFactory, BaseDNSServer
from .dns_upstream import DNSUpstreamServerMixin


CONTENT_FORMAT_DNS_MESSAGE = 553


class NotAcceptable(aiocoap.error.ConstructionRenderableError):
    """Exception to represent a 4.06 Not Acceptable CoAP error response"""

    code = NOT_ACCEPTABLE


class DNSOverCoAPServerFactory(BaseServerFactory):
    """Factory to create DNS over CoAP servers"""

    # pylint: disable=too-few-public-methods

    class _InnerDNSUpstream(DNSUpstreamServerMixin):
        def __init__(self, dns_upstream):
            super().__init__(dns_upstream=dns_upstream)
            self.dns_upstream = dns_upstream
            self._pending_responses = {}

        def send_response_to_requester(self, response, requester):
            assert requester in self._pending_responses
            self._pending_responses[requester].set_result(response)

        # pylint: disable=missing-function-docstring
        async def dns_response(self, query):
            self._pending_responses[query] = asyncio.Future()
            self.dns_query_received(query, query)
            return await self._pending_responses[query]

    class DNSQueryResource(aiocoap.resource.Resource):
        """The DNS over CoAP resource of the DNS over CoAP server.

        :param factory: The factory that created the DNS over CoAP server.
        :type factory: :py:class:`DNSOverCoAPServerFactory`
        """

        def __init__(self, factory):
            super().__init__()
            self._dns_upstream = factory._InnerDNSUpstream(factory.dns_upstream)

        @staticmethod
        def _coap_response(dns_response):
            return aiocoap.Message(
                content_format=CONTENT_FORMAT_DNS_MESSAGE,  # noqa: E501
                payload=dns_response,
            )

        async def _render_acceptable(self, request, query):
            formatted_response = await self._dns_upstream.dns_response(query)
            if (
                request.opt.accept is None
                or request.opt.accept == CONTENT_FORMAT_DNS_MESSAGE
            ):
                dns_response = formatted_response
            else:
                raise NotAcceptable()
            return self._coap_response(dns_response)

        async def _render_with_payload(self, request):
            if request.opt.content_format == CONTENT_FORMAT_DNS_MESSAGE:
                query = request.payload
            else:
                raise aiocoap.error.UnsupportedContentFormat()
            return await self._render_acceptable(request, query)

        async def render_fetch(self, request):
            """:py:class:`aiocoap.resource.Resource` method to serve a FETCH
            request

            :param request: The FETCH request
            :returns: The response for the FETCH request.
            """
            return await self._render_with_payload(request)

        async def render_get(self, request):
            """:py:class:`aiocoap.resource.Resource` method to serve a GET
            request

            :param request: The GET request
            :returns: The response for the GET request.
            """
            queries = dict(q.split("=") for q in request.opt.uri_query)
            if "dns" in queries:
                query_code = queries["dns"]
                # python's base64 only understands encoding with padding, so
                # add '=' padding if needed
                query = base64.urlsafe_b64decode(
                    query_code + "=" * (4 - len(query_code) % 4)
                )
                return await self._render_acceptable(request, query)
            raise aiocoap.error.BadRequest()

        async def render_post(self, request):
            """:py:class:`aiocoap.resource.Resource` method to serve a POST
            request

            :param request: The POST request
            :returns: The response for the POST request.
            """
            return await self._render_with_payload(request)

    class ClosableContext(aiocoap.Context, BaseDNSServer):
        """:py:class:`aiocoap.Context` that also serves as an extension of
        :py:class:`.dns_server.BaseDNSServer` so it can be returned by
        :py:meth:`DNSOverCoAPServerFactory.create_server`"""

        async def close(self):
            ri_type = type(self.request_interfaces)
            if self.request_interfaces:
                try:
                    await self.shutdown()
                except AttributeError:  # pragma: no cover
                    # apparently one should not call shutdown twice ;-)
                    pass
                self.request_interfaces = ri_type()

    async def create_server(self, loop, *args, local_addr=None, **kwargs):
        """Creates an ``aiocoap`` server context.

        :param loop: the asyncio event loop the server should run in
        :type loop: :py:class:`asyncio.AbstractEventLoop`
        :param local_addr: A tuple for the created server to bind to. The first
                           element is the host part, the second element the
                           port.
        :type local_addr: :py:class:`typing.Tuple[str, int]`

        :returns: An :py:class:`ClosableContext` object representing an
                  ``aiocoap`` server context.
        :rtype: :py:class:`ClosableContext`
        """
        config = Config()
        if local_addr is None:
            local_addr = ("localhost", None)
        site = aiocoap.resource.Site()
        site.add_resource(
            [".well-known", "core"],
            aiocoap.resource.WKCResource(site.get_resources_as_linkheader),
        )
        path = (
            config.get("transports", {})
            .get("coap", {})
            .get("path", "dns")
            .strip("/")
            .split("/")
        )
        site.add_resource(path, self.DNSQueryResource(self))

        try:
            client_identity = config["dtls_credentials"]["client_identity"]
            psk = config["dtls_credentials"]["psk"]
        except KeyError as exc:
            raise RuntimeError(f"DTLS credential option {exc} not found") from exc

        os.environ["AIOCOAP_DTLSSERVER_ENABLED"] = "1"

        # pylint: disable=protected-access
        aiocoap.transports.tinydtls_server._SEND_SLEEP_WORKAROUND = config.get(
            "dtls", {}
        ).get("server_hello_done_delay", 0.0)
        ctx = await self.ClosableContext.create_server_context(site, local_addr)
        ctx.server_credentials.load_from_dict(
            {
                ":client": {
                    "dtls": {
                        "client-identity": {"ascii": client_identity},
                        "psk": {"ascii": psk},
                    }
                }
            }
        )
        return ctx



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

"""Base definitions for the serving side of the DNS proxy"""

import abc
import asyncio
import typing

from .dns_upstream import DNSUpstream


class BaseDNSServer(abc.ABC):
    """An abstract DNS server."""

    # pylint: disable=too-few-public-methods
    @abc.abstractmethod
    async def close(self) -> typing.NoReturn:
        """Closes the server."""
        raise NotImplementedError


class BaseServerFactory(abc.ABC):
    """Abstract :py:class:`BaseDNSServer` factory.

    :param dns_upstream: The proxied DNS server for
                         :py:class:`.dns_upstream.DNSUpstreamServerMixin`.
    :type upstream_host: :py:class:`.dns_upstream.DNSUpstream`
    """  # noqa: E501

    # pylint: disable=too-few-public-methods
    def __init__(self, dns_upstream: DNSUpstream):
        self.dns_upstream = dns_upstream

    @abc.abstractmethod
    async def create_server(
        self,
        loop: asyncio.AbstractEventLoop,
        *args,
        local_addr: typing.Optional[typing.Tuple[str, int]] = None,
        **kwargs
    ) -> BaseDNSServer:
        """Creates a :py:class:`BaseDNSServer` object.

        :param loop: the asyncio event loop the server should run in
        :type loop: :py:class:`asyncio.AbstractEventLoop`
        :param local_addr: A tuple for the created server to bind to. The first
                           element is the host part, the second element the
                           port. If ``local_addr`` is ``None`` or any of its
                           elements are ``local_addr`` are ``None``, a sensible
                           default is selected by the implementation.
        :type local_addr: :py:class:`typing.Tuple[str, int]`

        :returns: An object based on the :py:class:`BaseDNSServer` class.
        :rtype: :py:class:`BaseDNSServer`
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

"""DNS over DTLS serving side of the proxy."""

import abc
import asyncio
import logging
import time
import typing

from DTLSSocket import dtls  # type: ignore

from .config import Config
from .dns_server import BaseServerFactory, BaseDNSServer
from .dns_upstream import DNSUpstreamServerMixin


logger = logging.getLogger()


class BaseDTLSWrapper(abc.ABC):
    """An abstract wrapper for a DTLS implementation

    :param transport: The datagram transport the datagrams should be encrypted
                      and decrypted for.
    :type transport: :py:class:`asyncio.DatagramTransport`
    """

    def __init__(self, transport: asyncio.DatagramTransport):
        self.transport = transport

    @abc.abstractmethod
    def is_connected(self, addr: typing.Any) -> bool:
        """Check if a session with ``addr`` was established.

        :param addr: A remote endpoint (implementation-specific)
        :returns: ``True``, when a session with ``addr`` is established,
                  ``False`` if not.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sessions(self) -> typing.Sequence:
        """Returns all currently established sessions.

        :returns: A sequence of (implementation-specific) remote endpoints with
                  which a session is established.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def connect(self, addr: typing.Any) -> typing.NoReturn:
        """Establish a session with ``addr``.

        :param addr: An (implementation-specific) remote endpoint
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self, addr: typing.Any) -> typing.NoReturn:
        """Closes a session with ``addr``.

        :param addr: An (implementation-specific) remote endpoint
        """
        raise NotImplementedError

    @abc.abstractmethod
    def handle_message(
        self, msg: bytes, addr: typing.Any
    ) -> typing.Tuple[bytes, typing.Any, bool]:
        """Handles a DTLS message that came over the datagram transport.

        :param msg: An incoming DTLS message.
        :type msg: bytes
        :param addr: The remote endpoint as served by the datagram transport.
        :returns: A 3-tuple, containing

                  1. The unencrypted message,
                  2. The (implementation-specific) remote endpoint the
                     message was received from, an
                  3. A boolean, indicating if the last message established a
                     session with the remote endpoint.

                  If ``msg`` was a control message, the first and second
                  elements will be ``None``.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, msg: bytes, addr: typing.Any) -> typing.NoReturn:
        """Send a ``msg`` encrypted to ``addr``

        :param msg: The message to encrypt via DTLS.
        :type msg: bytes
        :param addr: An (implementation-specific) remote endpoint to send the
                     encrypted message to.
        """
        raise NotImplementedError


class TinyDTLSWrapper(BaseDTLSWrapper):
    """A wrapper for
    `tinydtls <https://projects.eclipse.org/projects/iot.tinydtls>`_.
    """

    EVENT_CONNECTED = 0x1DE
    _CT_HANDSHAKE = 22
    _HT_SERVER_HELLO_DONE = 14
    LOG_LEVEL = {
        # pylint: disable=c-extension-no-member
        logging.DEBUG: dtls.DTLS_LOG_DEBUG,
        logging.INFO: dtls.DTLS_LOG_INFO,
        logging.WARNING: dtls.DTLS_LOG_WARN,
        logging.ERROR: dtls.DTLS_LOG_CRIT,
        logging.CRITICAL: dtls.DTLS_LOG_EMERG,
    }

    def __init__(self, transport):
        super().__init__(transport)
        # pylint: disable=c-extension-no-member
        credentials = Config()["dtls_credentials"]
        client_identity = credentials["client_identity"].encode()
        psk = credentials["psk"].encode()
        self._dtls = dtls.DTLS(
            read=self._read,
            write=self._write,
            event=self._event,
            pskId=client_identity,
            pskStore={client_identity: psk},
        )
        dtls.setLogLevel(self.LOG_LEVEL[logger.level])
        self._active_sessions = {}
        self._app_data = None
        self._last_event = None

    def __del__(self):
        self._active_sessions.clear()
        del self._dtls
        self._dtls = None

    def _read(self, addr, data):
        self._app_data = (data, addr + (0, 0))
        return len(data)

    def _write(self, addr, data):
        if (
            len(data) > 13
            and data[0] == self._CT_HANDSHAKE
            and data[13] == self._HT_SERVER_HELLO_DONE
        ):
            delay = Config().get("dtls", {}).get("server_hello_done_delay")

            if delay:
                # Delay Server Hello Done for a bit
                time.sleep(delay)
        self.transport.sendto(data, addr + (0, 0))
        return len(data)

    def _event(self, level, code):
        # pylint: disable=unused-argument
        self._last_event = code

    def is_connected(self, addr):
        return addr in self._active_sessions

    def sessions(self):
        return list(self._active_sessions)

    def connect(self, addr):
        connection = self._dtls.connect(*addr)
        if connection:
            self._active_sessions[addr] = connection

    def close(self, addr):
        if self.is_connected(addr):
            self._dtls.close(self._active_sessions[addr])
            del self._active_sessions[addr]

    def handle_message(self, msg, addr):
        connected = False
        if isinstance(addr, tuple):
            ret = self._dtls.handleMessageAddr(addr[0], addr[1], msg)
        elif isinstance(addr, dtls.Session):  # pylint: disable=I1101
            ret = self._dtls.handleMessage(addr, msg)
            addr = addr.addr, addr.port, addr.flowinfo, addr.scope_id
        else:
            raise ValueError(f"Unexpected session type {type(addr)}")
        if ret < 0:
            logger.warning("Unable to handle incoming DTLS message from %s", addr)
            return None, None, connected
        if self._last_event == self.EVENT_CONNECTED and not self.is_connected(addr):
            # pylint: disable=c-extension-no-member
            self._active_sessions[addr] = dtls.Session(*addr[:4])
            connected = True
        self._last_event = None
        if self._app_data is None:
            logger.debug("Unable to fetch application data from DTLS")
            return None, None, connected
        data, addr = self._app_data
        self._app_data = None
        return data, addr, connected

    def write(self, msg, addr):
        if isinstance(addr, tuple):
            if not self.is_connected(addr):
                logger.warning("%s does not have an active session", addr)
                return
            addr = self._active_sessions[addr]
        self._dtls.write(addr, msg)


class DNSOverDTLSServerFactory(BaseServerFactory):
    """Factory to create DNS over DLTS servers"""

    # pylint: disable=too-few-public-methods
    DODTLS_PORT = 853
    dtls_class = TinyDTLSWrapper

    class DNSOverDTLSServer(BaseDNSServer, DNSUpstreamServerMixin):
        """DNS over DTLS server implementation.

        :param factory: The factory that created the DNS over DTLS server.
        :type factory: :py:class:`DNSOverDTLSServerFactory`
        """

        def __init__(self, factory):
            super().__init__(dns_upstream=factory.dns_upstream)
            self.factory = factory
            self.transport = None
            self._dtls = None

        def __del__(self):
            del self._dtls
            self._dtls = None

        def connection_made(self, transport):
            # pylint: disable=line-too-long
            """See `connection_made()`_

            .. _`connection_made()`: https://docs.python.org/3/library/asyncio-protocol.html#asyncio.BaseProtocol.connection_made
            """  # noqa: E501
            self.transport = transport
            self._dtls = self.factory.dtls_class(self.transport)

        def datagram_received(self, data, addr):
            # pylint: disable=line-too-long
            """See `datagram_received()`_

            .. _`datagram_received()`: https://docs.python.org/3/library/asyncio-protocol.html#asyncio.DatagramProtocol.datagram_received
            """  # noqa: E501
            data, addr, _ = self._dtls.handle_message(data, addr)
            if data is None:
                return
            self.dns_query_received(data, addr)

        def error_received(self, exc):
            # pylint: disable=line-too-long
            """See `error_received()`_

            .. _`error_received()`: https://docs.python.org/3/library/asyncio-protocol.html#asyncio.DatagramProtocol.error_received
            """  # noqa: E501
            self._dtls = None  # pragma: no cover
            raise exc  # pragma: no cover

        def send_response_to_requester(self, response, requester):
            self._dtls.write(response, requester)

        async def close(self):
            if self.transport is not None:
                if self._dtls is not None:  # pragma: no cover
                    for session in self._dtls.sessions():
                        self._dtls.close(session)
                self.transport.close()
                self.transport = None

        def connection_lost(self, exc):  # pylint: disable=unused-argument
            # pylint: disable=line-too-long
            """See `connection_lost()`_

            .. _`connection_lost()`: https://docs.python.org/3/library/asyncio-protocol.html#asyncio.BaseProtocol.connection_lost
            """  # noqa: E501
            self._dtls = None

    def _create_server_protocol(self, *args, **kwargs):
        try:
            config = Config()
            _ = config["dtls_credentials"]["client_identity"]
            _ = config["dtls_credentials"]["psk"]
        except KeyError as exc:
            raise RuntimeError(f"DTLS credential option {exc} not found") from exc
        return self.DNSOverDTLSServer(self, *args, **kwargs)

    async def create_server(self, loop, *args, local_addr=None, **kwargs):
        """Creates an :py:class:`DNSOverDTLSServer` object.

        :param loop: the asyncio event loop the server should run in
        :type loop: :py:class:`asyncio.AbstractEventLoop`
        :param local_addr: A tuple for the created server to bind to. The first
                           element is the host part, the second element the
                           port.
        :type local_addr: :py:class:`typing.Tuple[str, int]`

        :returns: An :py:class:`DNSOverDTLSServer` object representing an
                  DNS over DTLS server.
        :rtype: :py:class:`DNSOverDTLSServer`
        """
        if local_addr is None:
            local_addr = ("localhost", self.DODTLS_PORT)
        if local_addr[1] is None:
            local_addr = (local_addr[0], self.DODTLS_PORT)

        _, protocol = await loop.create_datagram_endpoint(
            self._create_server_protocol,
            *args,
            local_addr=local_addr,
            **kwargs,
        )
        return protocol



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

# pylint: disable=missing-module-docstring

__version__ = "v0.0.7b2"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

import asyncio
import logging

import dns.message
import pytest

from DTLSSocket import dtls as tinydtls

from aiodnsprox import dns_upstream
from aiodnsprox import dtls

from .fixtures import dns_server, config  # noqa: F401


def test_tinydtls_wrapper__connect(mocker, config):
    config.add_config(
        {
            "dtls_credentials": {
                "client_identity": "Client_identifier",
                "psk": "secretPSK",
            }
        }
    )
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._read")
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._write")
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._event")
    wrapper = dtls.TinyDTLSWrapper(mocker.MagicMock())
    wrapper.connect(("::1", 853))
    wrapper._write.assert_called()
    wrapper.close(("::1", 853))


def test_tinydtls_wrapper__handle_message(caplog, mocker, config):
    config.add_config(
        {
            "dtls_credentials": {
                "client_identity": "Client_identifier",
                "psk": "secretPSK",
            }
        }
    )
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._read")
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._write")
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._event")
    wrapper = dtls.TinyDTLSWrapper(mocker.MagicMock())
    assert wrapper.handle_message(b"abcd", ("::1", 853)) == (None, None, False)
    assert wrapper.handle_message(b"abcd", tinydtls.Session("::1", 853)) == (
        None,
        None,
        False,
    )
    with pytest.raises(ValueError):
        wrapper.handle_message(b"abcd", "This is a string were non should be")
    with caplog.at_level(logging.WARNING):
        # try to handle Client Hello from unverified peer
        assert wrapper.handle_message(
            b"\x16\xfe\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x36\x01"
            b"\x00\x00\x2a\x00\x00\x00\x00\x00\x00\x00\x2a\xfe\xfd\x00"
            b"\x00\x00\x00\xc1\xfa\x0e\x7b\xbb\x9b\x30\xc8\xf2\xf0\x65"
            b"\xe7\x6b\x15\x59\x68\xea\x69\x30\xb6\x08\x6e\x58\xba\x1e"
            b"\x99\x61\x55\x00\x00\x00\x02\xc0\xa8\x01\x00",
            ("::1", 853),
        ) == (None, None, False)
    # Seems not to be logged like that anymore
    # assert "Unable to handle incoming DTLS message from ('::1', 853)" in caplog.text


def test_tinydtls_wrapper__write(caplog, mocker, config):
    config.add_config(
        {
            "dtls_credentials": {
                "client_identity": "Client_identifier",
                "psk": "secretPSK",
            }
        }
    )
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._read")
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._write")
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper._event")
    wrapper = dtls.TinyDTLSWrapper(mocker.MagicMock())
    with caplog.at_level(logging.WARNING):
        wrapper.write(b"abcd", ("::1", 853))
    assert "('::1', 853) does not have an active session" in caplog.text
    # pylint: disable=no-member,protected-access
    wrapper._write.assert_not_called()
    mocker.patch("aiodnsprox.dtls.TinyDTLSWrapper.is_connected", return_value=True)
    wrapper._active_sessions = {("::1", 853): tinydtls.Session("::1", 853)}
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        wrapper.write(b"abcd", ("::1", 853))
    assert "('::1', 853) does not have an active session" not in caplog.text
    wrapper._write.assert_called()
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        wrapper.write(b"abcd", wrapper._active_sessions["::1", 853])
    assert "('::1', 853) does not have an active session" not in caplog.text
    wrapper._write.assert_called()


@pytest.mark.asyncio
@pytest.mark.parametrize("done_delay", [None, 0.0, 0.012])
async def test_dtls_proxy(dns_server, config, done_delay):  # noqa: C901, F811
    proxy_bind = ("::1", 2304)
    config.add_config(
        {
            "dtls_credentials": {
                "client_identity": "Client_identifier",
                "psk": "secretPSK",
            }
        }
    )

    if done_delay is not None:
        config.add_config(
            {
                "dtls": {"server_hello_done_delay": done_delay},
            }
        )

    class ClientProtocol(asyncio.DatagramProtocol):
        def __init__(self, loop, on_connection_lost):
            self.loop = loop
            self.on_connection_lost = on_connection_lost
            self.connection = None
            self.result = None
            self.transport = None
            self._dtls = None

        def send_query(self, session):
            query = dns.message.make_query(
                dns_server["req_hostname"], dns_server["resp_rtype"].__name__
            )
            self._dtls.write(query.to_wire(), session)

        def connection_made(self, transport):
            self.transport = transport
            self._dtls = dtls.TinyDTLSWrapper(transport)
            self.connection = self._dtls.connect(proxy_bind)

        def datagram_received(self, response, session):
            data, addr, connected = self._dtls.handle_message(response, session)
            # during handle message the client connected
            if connected:
                self.send_query(session)
                return
            if data is not None:
                self.result = (dns.message.from_wire(data), addr)
                self.transport.close()

        def error_received(self, exc):
            raise exc

        def connection_lost(self, exc):
            self.on_connection_lost.set_result(self.result)
            self._dtls.close(proxy_bind)

    upstream = dns_upstream.DNSUpstream(dns_server["host"], dns_server["port"])
    factory = dtls.DNSOverDTLSServerFactory(upstream)
    loop = asyncio.get_running_loop()
    server = await factory.create_server(loop, local_addr=proxy_bind)
    try:
        on_connection_lost = loop.create_future()
        client_transport, _ = await loop.create_datagram_endpoint(
            lambda: ClientProtocol(loop, on_connection_lost), remote_addr=proxy_bind
        )
        try:
            response, addr = await on_connection_lost
            assert addr[:2] == (proxy_bind[0], proxy_bind[1])
            found_answer = False
            for rset in response.answer:
                for rda in rset:
                    if isinstance(rda, dns_server["resp_rtype"]):
                        found_answer = True
                        assert rda.address == dns_server["resp_address"]
            assert found_answer
        finally:
            client_transport.close()
    finally:
        await server.close()
        await server.close()  # call second time to check idempotency


@pytest.mark.asyncio
async def test_dtls_factory_create_server_wo_config(dns_server):
    proxy_bind = ("::1", 2304)
    upstream = dns_upstream.DNSUpstream(dns_server["host"], dns_server["port"])
    factory = dtls.DNSOverDTLSServerFactory(upstream)
    loop = asyncio.get_running_loop()
    with pytest.raises(RuntimeError):
        await factory.create_server(loop, local_addr=proxy_bind)


@pytest.mark.asyncio
@pytest.mark.parametrize("local_addr", [None, ("localhost", None)])
async def test_dtls_factory_create_server(local_addr, mocker):
    loop = mocker.MagicMock()
    future = asyncio.Future()
    future.set_result((0, 0))
    loop.create_datagram_endpoint.return_value = future
    upstream = dns_upstream.DNSUpstream("localhost", 53)
    factory = dtls.DNSOverDTLSServerFactory(upstream)
    server = await factory.create_server(loop, local_addr=local_addr)
    loop.create_datagram_endpoint.assert_called_once()
    del server



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

"""Proxy configuration"""

import argparse
import pprint
import typing
import yaml


class _Singleton(type):
    """Singleton meta class.

    see https://stackoverflow.com/q/6760685
    """

    _instances: dict[typing.Type["_Singleton"], "_Singleton"] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(metaclass=_Singleton):
    """Singleton Config mapping class to represent both config file and CLI
    argument congfiguration.
    """

    def __init__(self):
        self._sections = {}

    def __str__(self):
        return pprint.pformat(self._sections)

    def __repr__(self):
        return f"<{type(self).__name__}: {self._sections}>"

    def __len__(self):
        return len(self._sections)

    def __contains__(self, key: str) -> bool:
        return key in self._sections

    def __getitem__(self, key: str) -> typing.Mapping:
        return self._sections[key]

    def get(self, section: str, default=None) -> typing.Mapping:
        """Get configuration section.

        :param key: Name of the configuration section.
        :type key: str
        :param default: (Optional) default if section does not exist.
        """
        return self._sections.get(section, default)

    def add_config(self, config: typing.Mapping):
        """Adds configuration from a mapping

        :param config: A mapping that contains the new configuration sections.
        :type config: :py:class:`typing.Mapping`.
        """
        self._sections.update(config)

    def add_yaml_config(self, yaml_file):
        """Adds configuration from a YAML file

        :param yaml: A file-like object to a YAML file.
        :type yaml: A file-like object.
        """
        self.add_config(yaml.load(yaml_file, yaml.loader.FullLoader))

    def add_args_config(self, args):
        """Adds configuration from CLI arguments

        :param args: parsed CLI arguments.
        :type args: :py:class:`argparse.Namespace`
        """
        self.add_config(
            {
                k: vars(v) if isinstance(v, argparse.Namespace) else v
                for k, v in vars(args).items()
                if v is not None
            }
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

import asyncio
import socket

import dns.message
import pytest

from aiodnsprox import dns_upstream
from aiodnsprox import udp

from .fixtures import dns_server  # noqa: F401


@pytest.mark.asyncio
async def test_udp_proxy(dns_server):  # noqa: C901, F811
    proxy_bind = ("::1", 59462)

    class ClientProtocol(asyncio.DatagramProtocol):
        def __init__(self, on_connection_lost):
            self.on_connection_lost = on_connection_lost
            self.result = None
            self.transport = None

        def connection_made(self, transport):
            query = dns.message.make_query(
                dns_server["req_hostname"], dns_server["resp_rtype"].__name__
            )
            self.transport = transport
            self.transport.sendto(query.to_wire())

        def datagram_received(self, response, addr):
            self.result = (dns.message.from_wire(response), addr)
            self.transport.close()

        def error_received(self, exc):
            raise exc

        def connection_lost(self, exc):
            self.on_connection_lost.set_result(self.result)

    upstream = dns_upstream.DNSUpstream(dns_server["host"], dns_server["port"])
    factory = udp.DNSOverUDPServerFactory(upstream)
    loop = asyncio.get_running_loop()
    server = await factory.create_server(loop, local_addr=proxy_bind)
    try:
        on_connection_lost = loop.create_future()
        client_transport, _ = await loop.create_datagram_endpoint(
            lambda: ClientProtocol(on_connection_lost), remote_addr=proxy_bind
        )
        try:
            response, addr = await on_connection_lost
            assert addr == (proxy_bind[0], proxy_bind[1], 0, 0)
            found_answer = False
            for rset in response.answer:
                for rda in rset:
                    if isinstance(rda, dns_server["resp_rtype"]):
                        found_answer = True
                        assert rda.address == dns_server["resp_address"]
            assert found_answer
        finally:
            client_transport.close()
    finally:
        await server.close()
        await server.close()  # call second time to check idempotency


@pytest.mark.asyncio
@pytest.mark.parametrize("local_addr", [None, ("localhost", None)])
async def test_udp_factory_create_server(local_addr, mocker):
    loop = mocker.MagicMock()
    future = asyncio.Future()
    future.set_result((0, 0))
    loop.create_datagram_endpoint.return_value = future
    upstream = dns_upstream.DNSUpstream("localhost", 53)
    factory = udp.DNSOverUDPServerFactory(upstream)
    server = await factory.create_server(loop, local_addr=local_addr)
    loop.create_datagram_endpoint.assert_called_once()
    del server



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

import os

from setuptools import setup, find_packages

PACKAGE = "aiodnsprox"
DESCRIPTION = "A Python-based DNS-over-X proxy based on aiocoap "
AUTHOR = "Martine S. Lenders"
AUTHOR_EMAIL = "martine.lenders@tu-dresden.de"
URL = "https://github.com/netd-tud/aiodnsprox"


def get_requirements():
    with open("requirements.txt") as req_file:
        for line in req_file:
            yield line.strip()


def get_version(package):
    """Extract package version without importing file
    Importing cause issues with coverage,
        (modules can be removed from sys.modules to prevent this)
    Importing __init__.py triggers importing rest and then requests too

    Inspired from pep8 setup.py
    """
    with open(os.path.join(package, "__init__.py")) as init_fd:
        for line in init_fd:
            if line.startswith("__version__"):
                return eval(line.split("=")[-1])  # pylint:disable=eval-used
    return None


setup(
    name=PACKAGE,
    version=get_version(PACKAGE),
    description=DESCRIPTION,
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    license="MIT",
    download_url=URL,
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Intended Audience :: End Users/Desktop",
        "Environment :: Console",
        "Topic :: Utilities",
    ],
    install_requires=list(get_requirements()),
    entry_points={
        "console_scripts": [
            "aiodns-proxy = aiodnsprox.cli.proxy:sync_main",
        ],
    },
    python_requires=">=3.7",
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

import dns.message
import dns.rcode
import pytest

from aiodnsprox import dns_upstream

from .fixtures import dns_server  # noqa: F401


def test_upstream_init__unknown_transport():
    with pytest.raises(ValueError):
        dns_upstream.DNSUpstream(host="::1", transport=57)


@pytest.mark.parametrize(
    "port, transport, exp_port",
    [
        (None, dns_upstream.DNSTransport.UDP, 53),
        (None, dns_upstream.DNSTransport.UDP_TCP_FALLBACK, 53),
        (None, dns_upstream.DNSTransport.TCP, 53),
        (52387, dns_upstream.DNSTransport.UDP, 52387),
        (52387, dns_upstream.DNSTransport.UDP_TCP_FALLBACK, 52387),
        (52387, dns_upstream.DNSTransport.TCP, 52387),
    ],
)
def test_upstream_init(port, transport, exp_port):
    upstream = dns_upstream.DNSUpstream(host="::1", port=port, transport=transport)
    assert upstream.port == exp_port


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "id, transport, timeout",
    [
        (0, dns_upstream.DNSTransport.UDP, None),
        (41905, dns_upstream.DNSTransport.UDP, None),
        (0, dns_upstream.DNSTransport.UDP_TCP_FALLBACK, None),
        (41905, dns_upstream.DNSTransport.UDP_TCP_FALLBACK, None),
        (0, dns_upstream.DNSTransport.TCP, None),
        (41905, dns_upstream.DNSTransport.TCP, None),
        (0, dns_upstream.DNSTransport.UDP, 6.0),
        (41905, dns_upstream.DNSTransport.UDP, 6.0),
        (0, dns_upstream.DNSTransport.UDP_TCP_FALLBACK, 6.0),
        (41905, dns_upstream.DNSTransport.UDP_TCP_FALLBACK, 6.0),
        (0, dns_upstream.DNSTransport.TCP, 6.0),
        (41905, dns_upstream.DNSTransport.TCP, 6.0),
    ],
)
async def test_upstream_query(dns_server, id, transport, timeout):  # noqa: F811, E501
    upstream = dns_upstream.DNSUpstream(
        host=dns_server["host"], port=dns_server["port"], transport=transport
    )
    query = dns.message.make_query(
        dns_server["req_hostname"], dns_server["resp_rtype"].__name__
    )
    query.id = id
    response_bytes = await upstream.query(query.to_wire(), timeout=timeout)
    response = dns.message.from_wire(response_bytes)
    found_answer = False
    for rset in response.answer:
        for rda in rset:
            if isinstance(rda, dns_server["resp_rtype"]):
                found_answer = True
                assert rda.address == dns_server["resp_address"]
    assert found_answer


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "transport",
    [
        (dns_upstream.DNSTransport.UDP),
        (dns_upstream.DNSTransport.UDP_TCP_FALLBACK),
        (dns_upstream.DNSTransport.TCP),
    ],
)
async def test_upstream_query_timeout(transport):
    upstream = dns_upstream.DNSUpstream(host="::1", port=13417, transport=transport)
    query = dns.message.make_query("example.org", "AAAA")
    response_bytes = await upstream.query(query.to_wire(), timeout=0.1)
    response = dns.message.from_wire(response_bytes)
    assert response.rcode() == dns.rcode.SERVFAIL


def test_mock_upstream_init__unknown_address_type():
    with pytest.raises(TypeError):
        dns_upstream.MockDNSUpstream(IN={"A": dns.message.Message()})
    with pytest.raises(TypeError):
        dns_upstream.MockDNSUpstream(IN={"AAAA": dns.message.Message()})


@pytest.mark.parametrize(
    "A, AAAA",
    [
        ("10.0.0", None),
        ("10.0.0.1.2", None),
        (None, "2001:db8::1::1"),
        ("2001:db8::1", "10.0.0.1"),
        (b"\x0a\0\0\1\3", None),
        (b"\x0a\0\0", None),
        (None, b"\x20\x01\x0d\xb8" + (b"\0" * 11) + b"\1\3"),
        (None, b"\x20\x01\x0d\xb8"),
    ],  # pylint: disable=invalid-name
)
def test_mock_upstream_init__invalid_addresses(A, AAAA):
    IN = {}  # pylint: disable=invalid-name
    if A is not None:
        IN["A"] = A
    if AAAA is not None:
        IN["AAAA"] = AAAA
    with pytest.raises(ValueError):
        dns_upstream.MockDNSUpstream(IN=IN)


@pytest.mark.parametrize(
    "A, AAAA",
    [
        (None, None),
        ("10.0.0.1", None),
        (None, "2001:db8::1"),
        ("10.0.0.1", "2001:db8::1"),
        (b"\x0a\0\0\1", None),
        (None, b"\x20\x01\x0d\xb8" + (b"\0" * 11) + b"\1"),
    ],  # pylint: disable=invalid-name
)
def test_mock_upstream_init__success(A, AAAA):
    IN = {}  # pylint: disable=invalid-name
    if A is not None:
        IN["A"] = A
    if AAAA is not None:
        IN["AAAA"] = AAAA
    upstream = dns_upstream.MockDNSUpstream(IN=IN)
    if A is not None:
        assert "A" in upstream._IN  # pylint: disable=protected-access
    if AAAA is not None:
        assert "AAAA" in upstream._IN  # pylint: disable=protected-access


@pytest.mark.asyncio
async def test_mock_upstream_query_a():
    IN = {"A": "10.0.0.1"}  # pylint: disable=invalid-name
    upstream = dns_upstream.MockDNSUpstream(IN=IN)
    found_answer = False
    query = dns.message.make_query("example.org", "A").to_wire()
    response = dns.message.from_wire(await upstream.query(query))
    for rset in response.answer:
        for rda in rset:
            if isinstance(rda, dns.rdtypes.IN.A.A):
                found_answer = True
                assert rda.address == IN["A"]
    assert found_answer


@pytest.mark.asyncio
async def test_mock_upstream_query_aaaa():
    IN = {"AAAA": "2001:db8::1"}  # pylint: disable=invalid-name
    upstream = dns_upstream.MockDNSUpstream(IN=IN)
    found_answer = False
    query = dns.message.make_query("example.org", "AAAA").to_wire()
    response = dns.message.from_wire(await upstream.query(query))
    for rset in response.answer:
        for rda in rset:
            if isinstance(rda, dns.rdtypes.IN.AAAA.AAAA):
                found_answer = True
                assert rda.address == IN["AAAA"]
    assert found_answer


@pytest.mark.asyncio
async def test_mock_upstream_query_a_and_aaa():
    IN = {  # pylint: disable=invalid-name
        "A": "10.0.0.1",
        "AAAA": "::1",
    }
    upstream = dns_upstream.MockDNSUpstream(IN=IN)
    found_answer = 0
    query = dns.message.make_query("example.org", "A")
    query.find_rrset(
        query.question,
        dns.name.from_text("v6.example.org"),
        dns.rdataclass.IN,
        dns.rdatatype.AAAA,
        create=True,
        force_unique=True,
    )
    response = dns.message.from_wire(await upstream.query(query.to_wire()))
    for rset in response.answer:
        for rda in rset:
            if isinstance(rda, dns.rdtypes.IN.A.A):
                found_answer += 1
                assert rda.address == IN["A"]
            if isinstance(rda, dns.rdtypes.IN.AAAA.AAAA):
                found_answer += 1
                assert rda.address == IN["AAAA"]
    assert found_answer == 2


@pytest.mark.asyncio
async def test_mock_upstream_query_cname():
    IN = {  # pylint: disable=invalid-name
        "A": "10.0.0.1",
        "AAAA": "::1",
    }
    upstream = dns_upstream.MockDNSUpstream(IN=IN)
    found_answer = 0
    query = dns.message.make_query("example.org", "CNAME")
    response = dns.message.from_wire(await upstream.query(query.to_wire()))
    for rset in response.answer:
        for rda in rset:
            if isinstance(rda, dns.rdtypes.IN.A.A):
                found_answer += 1
            if isinstance(rda, dns.rdtypes.IN.AAAA.AAAA):
                found_answer += 1
    assert found_answer == 0


@pytest.mark.asyncio
async def test_mock_upstream_query_non_in():
    IN = {  # pylint: disable=invalid-name
        "A": "10.0.0.1",
        "AAAA": "::1",
    }
    upstream = dns_upstream.MockDNSUpstream(IN=IN)
    found_answer = 0
    query = dns.message.make_query("example.org", "CNAME", rdclass="CHAOS")
    response = dns.message.from_wire(await upstream.query(query.to_wire()))
    for rset in response.answer:
        for rda in rset:
            if isinstance(rda, dns.rdtypes.IN.A.A):
                found_answer += 1
            if isinstance(rda, dns.rdtypes.IN.AAAA.AAAA):
                found_answer += 1
    assert found_answer == 0



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (C) 2021-23 Freie Universität Berlin
# Copyright (C) 2023-25 TU Dresden
#
# Distributed under terms of the MIT license.

import asyncio

import aiocoap
import base64
import dns.message
import pytest

from aiodnsprox import coap
from aiodnsprox import dns_upstream

from .fixtures import dns_server, config


async def _client(proxy_bind, config):
    client = await aiocoap.Context.create_client_context()
    client.client_credentials.load_from_dict(
        {
            f"coaps://[{proxy_bind[0]}]/*": {
                "dtls": {
                    "client-identity": {
                        "ascii": config["dtls_credentials"]["client_identity"],
                    },
                    "psk": {
                        "ascii": config["dtls_credentials"]["psk"],
                    },
                },
            },
        }
    )
    return client


async def _send_query(
    client,
    uri,
    method,
    query,
    accept=coap.CONTENT_FORMAT_DNS_MESSAGE,
    content_format=coap.CONTENT_FORMAT_DNS_MESSAGE,
):
    request = aiocoap.Message(
        code=method,
        uri=uri,
        accept=accept,
        content_format=content_format,
        payload=query,
    )
    return await client.request(request).response


async def _send_query_safe(
    client,
    uri,
    method,
    query,
    accept=coap.CONTENT_FORMAT_DNS_MESSAGE,
    content_format=coap.CONTENT_FORMAT_DNS_MESSAGE,
    dns_var="dns",
):
    if method == aiocoap.GET:
        query_code = base64.urlsafe_b64encode(query).decode().strip("=")
        uri = f"{uri}?{dns_var}={query_code}"
    return await _send_query(
        client, uri, method, query, accept=accept, content_format=content_format
    )


async def _test_dns_query(client, dns_server, uri, method, query):
    coap_response = await _send_query_safe(client, uri, method, query)
    if method == aiocoap.POST:
        assert coap_response.code == aiocoap.numbers.codes.Code.CHANGED
    else:
        assert coap_response.code == aiocoap.numbers.codes.Code.CONTENT
    dns_response = dns.message.from_wire(coap_response.payload)
    found_answer = False
    for rset in dns_response.answer:
        for rda in rset:
            if isinstance(rda, dns_server["resp_rtype"]):
                found_answer = True
                assert rda.address == dns_server["resp_address"]
    assert found_answer


@pytest.mark.asyncio
@pytest.mark.parametrize("path", [None, "/foobar"])
async def test_coap_proxy(dns_server, config, path):
    proxy_bind = ("::1", None)
    config.add_config(
        {
            "dtls_credentials": {
                "client_identity": "Client_identifier",
                "psk": "secretPSK",
            }
        }
    )
    if path is None:
        path = "/dns"
    else:
        config.add_config({"transports": {"coap": {"path": path}}})

    upstream = dns_upstream.DNSUpstream(dns_server["host"], dns_server["port"])
    factory = coap.DNSOverCoAPServerFactory(upstream)
    loop = asyncio.get_running_loop()
    server = await factory.create_server(loop, local_addr=proxy_bind)
    try:
        client = await _client(proxy_bind, config)
        query = dns.message.make_query(
            dns_server["req_hostname"], dns_server["resp_rtype"].__name__
        )
        query = query.to_wire()
        await _test_dns_query(
            client,
            dns_server,
            f"coap://[{proxy_bind[0]}]{path}",
            aiocoap.FETCH,
            query,
        )
        await _test_dns_query(
            client,
            dns_server,
            f"coaps://[{proxy_bind[0]}]{path}",
            aiocoap.FETCH,
            query,
        )
        await _test_dns_query(
            client,
            dns_server,
            f"coap://[{proxy_bind[0]}]{path}",
            aiocoap.GET,
            query,
        )
        await _test_dns_query(
            client,
            dns_server,
            f"coaps://[{proxy_bind[0]}]{path}",
            aiocoap.GET,
            query,
        )
        await _test_dns_query(
            client,
            dns_server,
            f"coap://[{proxy_bind[0]}]{path}",
            aiocoap.POST,
            query,
        )
        await _test_dns_query(
            client,
            dns_server,
            f"coaps://[{proxy_bind[0]}]{path}",
            aiocoap.POST,
            query,
        )
        coap_response = await _send_query_safe(
            client,
            f"coaps://[{proxy_bind[0]}]{path}",
            aiocoap.FETCH,
            query,
            content_format=0,
        )
        assert (
            coap_response.code == aiocoap.numbers.codes.Code.UNSUPPORTED_CONTENT_FORMAT
        )
        coap_response = await _send_query_safe(
            client,
            f"coaps://[{proxy_bind[0]}]{path}",
            aiocoap.FETCH,
            query,
            accept=0,
        )
        assert coap_response.code == aiocoap.numbers.codes.Code.NOT_ACCEPTABLE
        coap_response = await _send_query_safe(
            client,
            f"coaps://[{proxy_bind[0]}]{path}",
            aiocoap.GET,
            query,
            dns_var="foobar",
        )
        assert coap_response.code == aiocoap.numbers.codes.Code.BAD_REQUEST
    finally:
        await server.close()
        await server.close()  # call second time to check idempotency


@pytest.mark.asyncio
async def test_coap_factory_create_server_wo_config(dns_server):
    proxy_bind = ("::1", 2304)
    upstream = dns_upstream.DNSUpstream(dns_server["host"], dns_server["port"])
    factory = coap.DNSOverCoAPServerFactory(upstream)
    loop = asyncio.get_running_loop()
    with pytest.raises(RuntimeError):
        await factory.create_server(loop, local_addr=proxy_bind)


@pytest.mark.asyncio
@pytest.mark.parametrize("local_addr", [None, ("localhost", None)])
async def test_coap_factory_create_server(local_addr, mocker, config):
    config.add_config(
        {
            "dtls_credentials": {
                "client_identity": "Client_identifier",
                "psk": "secretPSK",
            }
        }
    )
    loop = mocker.MagicMock()
    mocker.patch("aiodnsprox.coap.DNSOverCoAPServerFactory.ClosableContext")
    fut = asyncio.Future()
    fut.set_result(mocker.MagicMock())
    coap.DNSOverCoAPServerFactory.ClosableContext.create_server_context.return_value = (
        fut
    )
    upstream = dns_upstream.DNSUpstream("localhost", 53)
    factory = coap.DNSOverCoAPServerFactory(upstream)
    server = await factory.create_server(loop, local_addr=local_addr)
    coap.DNSOverCoAPServerFactory.ClosableContext.create_server_context.assert_called_once()
    del server
